import struct
from .common import NdrCommon
from ..tools import read_string_bindings
from ..uuid import bin_to_str, CLSID_SZ


class ScmReplyInfoData(NdrCommon):

    FMT1_32 = '<LLQL'
    FMT1_32_SZ = struct.calcsize(FMT1_32)

    FMT1_64 = '<LQQQ'
    FMT1_64_SZ = struct.calcsize(FMT1_64)

    FMT2 = '<LHHLHH'
    FMT2_SZ = struct.calcsize(FMT2)

    def __init__(self, data: bytes):
        offset = 0
        self.read_common(data, offset)
        offset += self.COMMON_SZ
        self.read_private(data, offset)
        offset += self.PRIVATE_SZ

        (
            pdw_reserverd,
            referent_id1,
            self.oxid,
            referent_id2,
        ) = struct.unpack_from(self.FMT1_32, data, offset=offset)
        offset += self.FMT1_32_SZ

        self.ipid_rem_unknown = bin_to_str(data, offset)
        offset += CLSID_SZ

        (
            self.authn_hint,
            version_major,
            version_minor,
            reserved,
            self.w_num_entries,
            self.w_security_offset,
        ) = struct.unpack_from(self.FMT2, data, offset=offset)
        offset += self.FMT2_SZ

        self.str_bindings, offset = read_string_bindings(data, offset=offset)



import struct
from .const import NTLMSSP_REVISION_W2K3


class NTLMVersion:

    VERSION_FMT = "<BBHBBBB"
    VERSION_SIZE = struct.calcsize(VERSION_FMT)

    def __init__(self, major, minor, build):
        """[MS-NLMP] 2.2.2.10 VERSION."""
        self.product_major_version = major
        self.product_minor_version = minor
        self.product_build = build
        self.reserved = 0  # must be ignored
        self.ntlm_revision_current = NTLMSSP_REVISION_W2K3

    def get_data(self) -> bytes:
        return struct.pack(
            self.VERSION_FMT,
            self.product_major_version,
            self.product_minor_version,
            self.product_build,
            self.reserved,
            self.reserved,
            self.reserved,
            self.ntlm_revision_current)



import hashlib
import string
import functools
from typing import Callable, Tuple
from .message_signature import NTLMMessageSignature
from .const import (
    NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY,
    NTLMSSP_NEGOTIATE_128,
    NTLMSSP_NEGOTIATE_56,
)


SIGN_CLIENT_SERVER = \
    b'session key to client-to-server signing key magic constant\x00'
SIGN_SERVER_CLIENT = \
    b'session key to server-to-client signing key magic constant\x00'


SEAL_CLIENT_SERVER = \
    b'session key to client-to-server sealing key magic constant\x00'
SEAL_SERVER_CLIENT = \
    b'session key to server-to-client sealing key magic constant\x00'


def sign_key(flags: int, random_session_key: bytes, client_mode: bool = True):
    if not (flags & NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY):
        return None

    md5 = hashlib.new('md5')
    if client_mode:
        md5.update(random_session_key + SIGN_CLIENT_SERVER)
    else:
        md5.update(random_session_key + SIGN_SERVER_CLIENT)
    return md5.digest()


def seal_key(flags: int, random_session_key: bytes, client_mode: bool = True):
    if flags & NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY:
        if flags & NTLMSSP_NEGOTIATE_128:
            seal_key = random_session_key
        elif flags & NTLMSSP_NEGOTIATE_56:
            seal_key = random_session_key[:7]
        else:
            seal_key = random_session_key[:5]

        md5 = hashlib.new('md5')
        if client_mode:
            md5.update(seal_key + SEAL_CLIENT_SERVER)
        else:
            md5.update(seal_key + SEAL_SERVER_CLIENT)
        return md5.digest()

    if flags & NTLMSSP_NEGOTIATE_56:
        return random_session_key[:7] + b'\xa0'

    return random_session_key[:5] + b'\xe5\x38\xb0'


def _seal(
        flags: int,
        seq_num: int,
        message_to_sign: bytes,
        message_to_encrypt: bytes,
        signing_key: bytes,
        handle: Callable) -> Tuple[bytes, bytes]:
    sealed_message = handle(message_to_encrypt)
    message_signature = NTLMMessageSignature.get_data(
        flags,
        seq_num,
        message_to_sign,
        signing_key,
        handle)
    return sealed_message, message_signature


def seal_func(
            signing_key: bytes,
            handle: Callable[[bytes, bytes], bytes],
        ) -> Callable[
            [int, int, bytes, bytes],
            Tuple[bytes, bytes]]:
    return functools.partial(
        _seal,
        signing_key=signing_key,
        handle=handle)


def sign_func(
            signing_key,
            handle: Callable
        ) -> Callable[[int, int, bytes], bytes]:
    return functools.partial(
        NTLMMessageSignature.get_data,
        signing_key=signing_key,
        handle=handle)



import logging
from typing import Union
from datetime import datetime, timedelta


_FMT = '%Y%m%d%H%M%S.%f%z'


def dt_from_str(s: str) -> Union[datetime, timedelta]:
    """String to datetime.

    It's crap, but at least citrix may return invalid values like:
        00010101000000.000000+060

    Therefore, we try and when we fail, this function will return
     datetime.fromtimestamp(0) or timedelta(0). If you enable debug logging,
    the library will generate a log line when such an exception occurs.

    https://www.dmtf.org/sites/default/files/standards/documents/DSP0004V2.3_final.pdf
    """  # nopep8
    s = s.replace('*', '0')

    if s.endswith(':000'):
        # ddddddddhhmmss.mmmmmm:000
        try:
            td = timedelta(
                days=int(s[:8]),
                hours=int(s[8:10]),
                minutes=int(s[10:12]),
                seconds=int(s[12:14]),
                microseconds=int(s[15:21])
            )
        except Exception as e:
            logging.debug(
                f'invalid interval `{s}` ({e}); return timedelta(0)')
            td = timedelta(0)
        return td

    try:
        # timestamp: yyyymmddhhmmss.mmmmmmsutc
        # utc is offset in minutes
        minutes = int(s[-3:])
        hours = minutes // 60
        minutes = minutes % 60

        for t in (4, 6):
            # months and days must start at 1, not 0
            e = t + 2
            if s[t:e] == '00':
                s = s[:t] + '01' + s[e:]

        dt = datetime.strptime(f"{s[:21]}{s[-4]}{hours:02}{minutes:02}", _FMT)

    except Exception as e:
        logging.debug(
            f'invalid datetime `{s}` ({e}); '
            'return datetime.fromtimestamp(0)')
        dt = datetime.fromtimestamp(0)

    return dt


if __name__ == '__main__':
    print(dt_from_str('20220207094949.500000+060'))
    print(dt_from_str('19980525133015.0000000-300'))
    print(dt_from_str('19980525183015.0000000+000'))
    print(dt_from_str('19980525******.0000000+000'))
    print(dt_from_str('1998**********.0000000+000'))
    print(dt_from_str('00010000000000.0000000+300'))
    print(dt_from_str('00000000000000.0000000+300'))
    print(dt_from_str('00000001132312.125***:000'))



import asyncio
import logging
from typing import TYPE_CHECKING
from .const import WBEM_INFINITE
from .const import WBEM_FLAG_FORWARD_ONLY
from .const import WBEM_FLAG_RETURN_IMMEDIATELY
from .dtypes.wordstr import WORDSTR
from .qcontext import QContext


if TYPE_CHECKING:
    from .protocol import Protocol
    from .connection import Connection


class Query:

    _WQL = WORDSTR('WQL')

    def __init__(
            self,
            query: str,
            namespace: str = 'root/cimv2',
            language: str = None):
        if not namespace.startswith('//'):
            namespace = '//./' + namespace
        self.query = query
        self.namespace = namespace
        self._query = WORDSTR(query)
        self._language = self._WQL if language is None else WORDSTR(language)

    def go(
            self,
            conn: 'Connection',
            proto: 'Protocol',
            flags: int = (
                WBEM_FLAG_RETURN_IMMEDIATELY | WBEM_FLAG_FORWARD_ONLY),
            timeout: int = 60,
            skip_optimize: bool = False):
        """IWbemServices_ExecQuery.
        3.1.4.3.18 IWbemServices::ExecQuery (Opnum 20) [MS-WMI]

        Optional flags: (import from aiowmi.const)

        WBEM_FLAG_USE_AMENDED_QUALIFIERS (0x00020000)
            If this bit is not set, the server SHOULD not return CIM
            localizable information.
            If this bit is set, the server SHOULD return CIM localizable
            information for the CIM object, as specified in section 2.2.6.
        WBEM_FLAG_RETURN_IMMEDIATELY (0x00000010)
            If this bit is not set, the server MUST make the method
            call synchronously. If this bit is set, the server MUST make the
            method call semisynchronously.
        WBEM_FLAG_DIRECT_READ (0x00000200)
            If this bit is not set, the server MUST consider the entire
            class hierarchy when it returns the result.
            If this bit is set, the server MUST disregard any derived
            class when it searches the result.
        WBEM_FLAG_PROTOTYPE (0x00000002)
            If this bit is not set, the server MUST run the query.
            If this bit is set, the server MUST only return the class
        WBEM_FLAG_FORWARD_ONLY (0x00000020)
            If this bit is not set, the server MUST return an
            enumerator that has reset capability.
            If this bit is set, the server MUST return an enumerator
            without reset capability, as specified in section
        """
        return QContext(self, conn, flags, proto, timeout, skip_optimize)

    async def start(
            self,
            conn: 'Connection',
            proto: 'Protocol',
            flags: int = (
                WBEM_FLAG_RETURN_IMMEDIATELY | WBEM_FLAG_FORWARD_ONLY),
            timeout: int = 60):
        self._qc = QContext(self, conn, flags, proto, timeout, True)
        await self._qc.start()

    def optimize(self) -> asyncio.Future:
        return self._qc.optimize()

    def next(self, timeout: int = WBEM_INFINITE) -> asyncio.Future:
        return self._qc.next()

    def done(self) -> asyncio.Future:
        qc, self._qc = self._qc, None
        return qc.release()



import struct
from ..tools import get_rangom_bytes, ntowf_v2, hmac_md5
from .av_pairs import AvPairs
from .version import NTLMVersion
from .const import NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED
from .const import NTLMSSP_NEGOTIATE_VERSION


# 2.2.1.3 AUTHENTICATE_MESSAGE
class NTLMAuthAuthenticate:
    """Message from client to server."""

    SIGNATURE = b'NTLMSSP\x00'
    SIGNATURE_LEN = len(SIGNATURE)
    MESSAGE_TYPE = 3
    AUTHENTICATE_FMT = '<LHHLHHLHHLHHLHHLHHLL'
    AUTHENTICATE_SIZE = SIGNATURE_LEN + struct.calcsize(AUTHENTICATE_FMT)

    def __init__(self, negotiate_flags):
        self._negotiate_flags = negotiate_flags
        self._version = b''
        self._version_len = 0
        self._mic = b''     # Message integrity, NTLMSSP_MESSAGE_SIGNATURE
        self._mic_len = 0   # Message integrity length is 16 when set
        self._lm_challenge_response = b''
        self._lm_challenge_response_len = 0
        self._nt_challenge_response = b''
        self._nt_challenge_response_len = 0
        self._domain_name = b''
        self._domain_name_len = 0
        self._user_name = b''
        self._user_name_len = 0
        self._workstation_name = b''
        self._workstation_name_len = 0
        self._encr_random_session_key = b''
        self._encr_random_session_key_len = 0

    def set_version(self, version: NTLMVersion):
        """Version is packed as 8 bytes.
        So:
          len(self._version) ==  NTLMVersion.VERSION_SIZE == 8 bytes
        """
        self._version = version.get_data()
        self._version_len = NTLMVersion.VERSION_SIZE
        self._negotiate_flags |= NTLMSSP_NEGOTIATE_VERSION

    def set_credentials(
            self,
            user_name: str,
            password: str,
            target_info: bytes,
            server_challenge: bytes,
            domain_name: str = '') -> bytes:
        self._user_name = user_name.encode('utf-16le')
        self._user_name_len = len(self._user_name)

        self._domain_name = domain_name.encode('utf-16le')
        self._domain_name_len = len(self._domain_name)

        response_key_nt = ntowf_v2(user_name, password, self._domain_name)

        av_pairs = AvPairs(target_info)
        av_pairs.set_target_name()
        av_time = av_pairs.get_or_set_av_time()

        client_challenge = get_rangom_bytes(8)
        server_name = av_pairs.get_data()

        temp = b''.join((
            b'\x01\x01\x00\x00\x00\x00\x00\x00',
            av_time,
            client_challenge,
            b'\x00\x00\x00\x00',
            server_name,
            b'\x00\x00\x00\x00'))

        nt_proof_str = hmac_md5(response_key_nt, server_challenge + temp)

        self._lm_challenge_response = hmac_md5(
            response_key_nt,
            server_challenge + client_challenge) + client_challenge
        self._nt_challenge_response = nt_proof_str + temp

        session_base_key = hmac_md5(response_key_nt, nt_proof_str)

        self._lm_challenge_response_len = len(self._lm_challenge_response)
        self._nt_challenge_response_len = len(self._nt_challenge_response)

        return session_base_key

    def set_workstation_name(self, workstation_name):
        self._workstation_name = workstation_name.encode('utf-16le')
        self._workstation_name_len = len(self._workstation_name)
        self._negotiate_flags |= NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED

    def set_encr_random_session_key(self, encr_random_session_key: bytes):
        self._encr_random_session_key = encr_random_session_key
        self._encr_random_session_key_len = len(self._encr_random_session_key)

    def get_data(self) -> bytes:
        domain_name_offset = \
            self.AUTHENTICATE_SIZE + self._version_len + self._mic_len
        user_name_offset = \
            domain_name_offset + self._domain_name_len
        workstation_name_offset = \
            user_name_offset + self._user_name_len
        lm_challenge_response_offset = \
            workstation_name_offset + self._workstation_name_len
        nt_challenge_response_offset = \
            lm_challenge_response_offset + self._lm_challenge_response_len
        encr_random_session_key_offset = \
            nt_challenge_response_offset + self._nt_challenge_response_len

        data = struct.pack(
            self.AUTHENTICATE_FMT,
            self.MESSAGE_TYPE,                  # <L
            self._lm_challenge_response_len,    # <H
            self._lm_challenge_response_len,    # <H  (max len, equal to len)
            lm_challenge_response_offset,       # <L
            self._nt_challenge_response_len,    # <H
            self._nt_challenge_response_len,    # <H  (max len, equal to len)
            nt_challenge_response_offset,       # <L
            self._domain_name_len,              # <H
            self._domain_name_len,              # <H  (max len, equal to len)
            domain_name_offset,                 # <L
            self._user_name_len,                # <H
            self._user_name_len,                # <H  (max len, equal to len)
            user_name_offset,                   # <L
            self._workstation_name_len,         # <H
            self._workstation_name_len,         # <H  (max len, equal to len)
            workstation_name_offset,            # <L
            self._encr_random_session_key_len,  # <H NTLMSSP_NEGOTIATE_KEY_EXCH
            self._encr_random_session_key_len,  # <H  (max len, equal to len)
            encr_random_session_key_offset,     # <L
            self._negotiate_flags,              # <L
        )
        data = b''.join([
            self.SIGNATURE,
            data,
            self._version,
            self._mic,
            self._domain_name,
            self._user_name,
            self._workstation_name,
            self._lm_challenge_response,
            self._nt_challenge_response,
            self._encr_random_session_key
        ])
        return data



import struct

# struct {
# p_cont_def_result_t result;

# p_provider_reason_t reason; /* only relevant if result !=
# * acceptance */
# p_syntax_id_t transfer_syntax;/* tr syntax selected
# * 0 if result not

# * accepted */
# } p_result_t;


class RpcResult:
    RESULT_LIST_FMT = '<BHH'  # result_n, reserved, reserved2
    RESULT_LIST_FMT_SIZE = struct.calcsize(RESULT_LIST_FMT)



"""OBJREF_CUSTUM

See: 2.2.18.6, OBJREF_CUSTOM
"""
import struct
from typing import Tuple
from ..uuid import bin_to_str, CLSID_SZ
from .objref import ObjRef


class ObjRefStd(ObjRef):

    __slots__ = (
        'std_flags',
        'c_public_refs',
        'oxid',
        'oid',
        'ipid')

    STANDARD_FMT = '<LLQQ'
    STANDARD_FMT_SZ = struct.calcsize(STANDARD_FMT)

    @classmethod
    def from_data(cls, data: bytes, offset: int) -> Tuple['ObjRefStd', int]:
        self = cls()

        (
            self.std_flags,
            self.c_public_refs,
            self.oxid,
            self.oid,
        ) = struct.unpack_from(cls.STANDARD_FMT, data, offset=offset)

        # 3.2.4.4 Managing Object Lifetime
        # if the public reference counter is 0, then we need to get a new
        # reference (RemAddRef) and release when finished (RemRelease).
        assert self.c_public_refs, 'public reference counter is 0'

        offset += cls.STANDARD_FMT_SZ

        self.ipid = bin_to_str(data, offset=offset)
        offset += CLSID_SZ

        return self, offset

    def get_data(self) -> bytes:
        return super().get_data() + struct.pack(
            self.STANDARD_FMT,
            self.std_flags,
            0,
            self.oxid,
            self.oid
        ) + self.sa_res_addr



import struct
from .com_version import COM_VERSION_MAJOR, COM_VERSION_MINOR
from ..tools import pad8, gen_referent_id
from .common import NdrCommon
from ..dcom_const import CLSID_InstantiationInfo


class InstantiationInfoData(NdrCommon):

    CLSID = CLSID_InstantiationInfo

    BUFFER32_FMT = '<LLlLLLLHHL'
    BUFFER32_SZ = struct.calcsize(BUFFER32_FMT)

    BUFFER64_FMT = '<LLlLLQLHHQ'
    BUFFER64_SZ = struct.calcsize(BUFFER64_FMT)

    def __init__(self, class_id: bytes, iid: bytes):
        self.class_id = class_id
        self.iid = [iid]

    def get_data(self) -> bytes:
        class_ctx = 0
        actv_flags = 0
        f_is_surrogate = 0
        c_idd = len(self.iid)
        inst_flag = 0
        referent_id = gen_referent_id()

        referents = b''.join(self.iid)

        this_size = \
            self.COMMON_SZ + \
            self.PRIVATE_SZ + \
            len(self.class_id) + \
            self.BUFFER32_SZ + \
            len(referents)

        # pre-calculate the padding to set this_size to the corrent value
        pad = pad8(this_size)
        this_size += pad

        buffer = self.class_id + struct.pack(
            self.BUFFER32_FMT,
            class_ctx,          # <L
            actv_flags,         # <L
            f_is_surrogate,     # <l (signed)
            c_idd,              # <L
            inst_flag,          # <L
            referent_id,        # <L (32) / <Q (64)
            this_size,          # <L this_size is set to 0 in impacket
            COM_VERSION_MAJOR,  # <H
            COM_VERSION_MINOR,  # <H
            len(self.iid),      # <L (32) / <Q (64) (N referents)
        ) + referents

        private, padding = self.get_private(buffer)

        return self.get_common() + private + buffer + padding



import struct
from typing import Tuple


class Heap:

    HEAP = '<L'
    HEAP_SZ = struct.calcsize(HEAP)

    @classmethod
    def from_data(cls, data: bytes, offset: int) -> Tuple[bytes, int]:
        # ClassHeap
        (
            heap_length,
        ) = struct.unpack_from(cls.HEAP, data, offset=offset)

        # HeapLength is a 32-bit value with the most significant bit always set
        # (using little-endian binary encoding for the 32-bit value), so that
        # the length is actually only 31 bits.
        heap_length &= 0x7fffffff
        offset += cls.HEAP_SZ

        heap = data[offset: offset + heap_length]
        offset += heap_length

        return heap, offset



import struct
from ..tools import gen_referent_id
from .activation_blob import ActivationBlob
from .activation_context_info_data import ActivationContextInfoData
from .activation_context_info_data import ActivationContextInfoData
from .instantiation_info_data import InstantiationInfoData
from .location_info_data import LocationInfoData
from .objref_custom import ObjRefCustom
from .orpcthis import ORPCTHIS
from .scm_request_info_data import ScmRequestInfoData


class RemoteCreateInstance:

    FMT32 = '<LLLL'
    FMT64 = '<QQLL'

    def __init__(self, class_id: bytes, iid: bytes):

        activatoin_blob = ActivationBlob()

        activatoin_blob.add_info_data(InstantiationInfoData(class_id, iid))
        activatoin_blob.add_info_data(LocationInfoData)
        activatoin_blob.add_info_data(ActivationContextInfoData)
        activatoin_blob.add_info_data(ScmRequestInfoData)

        obj_ref_custom = ObjRefCustom.init()
        obj_ref_custom.set_object(activatoin_blob.get_data())

        self.ab_data = obj_ref_custom.get_data()

    def get_data(self) -> bytes:

        p_unk_outer = 0
        referent_id = gen_referent_id()
        c_ab_data = len(self.ab_data)

        data = ORPCTHIS.get_data(flags=1) + struct.pack(
            self.FMT32,
            p_unk_outer,
            referent_id,
            c_ab_data,
            c_ab_data,
        ) + self.ab_data

        return data



import struct
from typing import Tuple

# typedef struct{
#   u_int8 [size_is(auth_pad_length)] auth_pad[]; /* align(4) */
#   u_int8 auth_type; /* :01 which authent service */
#   u_int8 auth_level; /* :01 which level within service */
#   u_int8 auth_pad_length; /* :01 */
#   u_int8 auth_reserved; /* :01 reserved, m.b.z. */
#   u_int32 auth_context_id; /* :04 */
#   u_int8 [size_is(auth_length)] auth_value[]; /* credentials */
# } auth_verifier_co_t;


class RpcAuthVerifierCo:
    __slots__ = (
        'auth_type',
        'auth_level',
        'auth_pad_length',
        'auth_context_id',
        'auth_value')

    AUTH_VERIFIER_CO_FMT = '<BBBBL'
    SIZE = struct.calcsize(AUTH_VERIFIER_CO_FMT)

    def __init__(self, data: bytes, auth_length: int, offset: int = 0):
        (
            self.auth_type,
            self.auth_level,
            self.auth_pad_length,
            reserved,
            self.auth_context_id,
        ) = struct.unpack_from(self.AUTH_VERIFIER_CO_FMT, data, offset)
        offset += self.SIZE
        self.auth_value = data[offset:offset+auth_length]

    @classmethod
    def make(
            cls,
            auth_type: int,
            auth_level: int,
            auth_pad_length: int,
            auth_context_id: int,
            auth_value: bytes) -> Tuple[bytes, int]:
        "Retruns the packet bytes and auth_length"
        padding = b'\xFF' * auth_pad_length

        return padding + struct.pack(
            cls.AUTH_VERIFIER_CO_FMT,
            auth_type,
            auth_level,
            auth_pad_length,
            0,  # reserved
            auth_context_id,
        ) + auth_value, len(auth_value)



"""
To Enctypt:
b'\x05\x00\x07\x00\x01\x00\x00....


To Sign:
Mayor, Minor, Ptype, Flags
\x05
\x00
\x00
\x03

Packed drep
\x10\x00\x00\x00

Frag len (512)
\x00\x02

Auth len
\x10\x00

CallId
\x02\x00\x00\x00

Alloc hint  (464, encrypted pduData)
\xd0\x01\x00\x00

ContextId
\x00\x00

Op num
\x04\x00

PduData
\x05\x00\x07\x00...

sec trailer
\n\x06\x00\x00\x7f5\x01\x00'
"""

import struct
from typing import TYPE_CHECKING, Optional
from .common import RpcCommon
from .cont_elem import RpcContElem
from .const import MSRPC_REQUEST, PFC_OBJECT_UUID
from ..rpc.auth_verifier_co import RpcAuthVerifierCo
from ..tools import pad4
from ..uuid import uuid_to_bin

if TYPE_CHECKING:
    from ..protocol import Protocol


class RpcRequest(RpcCommon):

    AUTH_SZ = 16
    AUTH_VALUE = b' '*AUTH_SZ

    def __init__(self, op_num: int, uuid_str: Optional[str] = None):
        super().init(MSRPC_REQUEST)
        if uuid_str is not None:
            self.pfc_flags |= PFC_OBJECT_UUID
        self.op_num = op_num
        self.uuid_str = uuid_str

    def seal_data(self, proto: 'Protocol', ctx_id: int = 0) -> bytes:
        auth_pad_length = pad4(self._pdu_data_len)
        message_to_encrypt = self._pdu_data
        alloc_hint = len(message_to_encrypt)
        pdu_data = struct.pack(
            '<LHH',
            alloc_hint,
            ctx_id,
            self.op_num,
        )
        if self.pfc_flags & PFC_OBJECT_UUID:
            uuid = uuid_to_bin(self.uuid_str)
            pdu_data += uuid

        self.set_pdu_data(pdu_data + message_to_encrypt)

        auth_verifier, auth_length = RpcAuthVerifierCo.make(
            proto._auth_type,
            proto._auth_level,
            auth_pad_length,
            4242,  # context id
            self.AUTH_VALUE)

        proto._dcom.set_call_id(self)
        self.set_auth_verifier(auth_verifier, auth_length)

        # need the auth verifier, but not the auth data
        message_to_sign = self.get_data()[:-self.AUTH_SZ]

        sealed_message, message_signature = proto._client_seal(
            flags=proto._flags,
            seq_num=proto._dcom.get_seq_num(),
            message_to_sign=message_to_sign,
            message_to_encrypt=message_to_encrypt)

        self.set_pdu_data(pdu_data + sealed_message)
        self.set_auth_data(message_signature)

        return self.get_data()

    def sign_data(self, proto: 'Protocol', ctx_id: int = 0) -> bytes:

        assert proto._flags == 3767042613

        auth_pad_length = pad4(self._pdu_data_len)

        alloc_hint = len(self._pdu_data)
        pdu_data = struct.pack(
            '<LHH',
            alloc_hint,
            ctx_id,
            self.op_num,
        )
        if self.pfc_flags & PFC_OBJECT_UUID:
            uuid = uuid_to_bin(self.uuid_str)
            pdu_data += uuid

        self.set_pdu_data(pdu_data + self._pdu_data)

        auth_verifier, auth_length = RpcAuthVerifierCo.make(
            proto._auth_type,
            proto._auth_level,
            auth_pad_length,
            4242,  # context id
            self.AUTH_VALUE)

        proto._dcom.set_call_id(self)
        self.set_auth_verifier(auth_verifier, auth_length)

        # need the auth verifier, but not the auth data
        message_to_sign = self.get_data()[:-self.AUTH_SZ]

        seq_num = proto._dcom.get_seq_num()

        message_signature = proto._client_sign(
            flags=proto._flags,
            seq_num=seq_num,
            message_to_sign=message_to_sign)

        self.set_auth_data(message_signature)

        return self.get_data()
